from __future__ import unicode_literals

from pycparser import parse_file
from pycparser.c_parser import CParser
from pycparser.c_ast import NodeVisitor

def get_supported_modules():
    return open('supported_modules.txt').read().split("\n")[:-1]

def _check_executable(executable):
    from distutils.spawn import find_executable
    return not(find_executable(executable) is None)

def _cpp_executable():
    if _check_executable('cpp'):
        return 'cpp'
    elif _check_executable('cl'):
        return 'cl'
    elif _check_executable('gcc'):
        return 'gcc'
    elif _check_executable('clang'):
        return 'clang'
    return 'cpp'

def _cpp_args():
    args = ['-Incephes/cephes']
    if _cpp_executable() == 'cl':
        args += ['/nologo', '/EP']
    return args

class FuncSign(object):
    def __init__(self, name, ret_type):
        self.name = name
        self.ret_type = ret_type
        self.param_names = []
        self.param_types = []

    def __str__(self):
        n = len(self.param_names)
        pnames = self.param_names
        ptypes = self.param_types

        sparams = [('%s %s' % (ptypes[i], pnames[i])) for i in range(n)]
        sparams = ', '.join(sparams)
        s = '%s %s(%s)' % (self.ret_type, self.name, sparams)
        return s


class FuncDefVisitor(NodeVisitor):

    def __init__(self):
        self.functions = []

    def _parse_param_signature(self, p):
        name = getattr(p.type, 'declname', '')

        suffix = ''
        t = p.type.type
        while not hasattr(t, 'names'):
            t = t.type
            suffix += '*'
        typ = t.names[0]
        return (name, typ + suffix)

    def visit_FuncDef(self, node):
        ret_type = node.decl.type.type.type.names[0]
        if len(node.decl.storage) == 0:
            fs = FuncSign(node.decl.name, ret_type)
            for p in node.decl.type.args.params:
                (name, typ) = self._parse_param_signature(p)
                fs.param_names.append(name)
                fs.param_types.append(typ)
            self.functions.append(fs)


def read_export_file(fp):
    lines = open(fp).read().split('\n')
    lines = [l for l in lines if len(l.strip()) > 0]
    d = dict()
    for line in lines:
        modname, funcnames = line.split(':')
        modname = modname.strip()
        d[modname] = [f.strip() for f in funcnames.split(',')]
    return d


def fetch_func_decl(filename):
    cpp_path = _cpp_executable()
    cpp_args = _cpp_args()
    ast = parse_file(filename, use_cpp=True, cpp_path=cpp_path,
                      cpp_args=cpp_args)

    v = FuncDefVisitor()
    v.visit(ast)

    return [str(f) for f in v.functions]


def _rcs(s):
    if s.startswith('cephes_'):
        return s[len('cephes_'):]
    return s


def api_fdecls(decl):
    parser = CParser()
    decl = parser.parse(decl, filename='<stdin>').ext[0]
    name = decl.name
    args = decl.type.args
    nargs = len(args.params)
    if len(decl.type.type.type.names) > 1:
        assert False
    else:
        rtype = decl.type.type.type.names[0]
    ndecl = rtype + ' ncephes_' + _rcs(name) + '('
    for param in args.params:
        if len(param.type.type.names) > 1:
            assert False
        typ = param.type.type.names[0]
        ndecl += typ + ' ' + param.name + ', '
    if nargs > 0:
        ndecl = ndecl[:-2]
    return ndecl + ');'


def forward_call(decl):
    parser = CParser()
    decl = parser.parse(decl, filename='<stdin>').ext[0]
    name = decl.name
    args = decl.type.args
    nargs = len(args.params)
    if len(decl.type.type.type.names) > 1:
        assert False
    else:
        rtype = decl.type.type.type.names[0]
    ndecl = rtype + ' ncephes_' + _rcs(name) + '('
    call_expr = name + '('
    for param in args.params:
        if len(param.type.type.names) > 1:
            assert False
        typ = param.type.type.names[0]
        ndecl += typ + ' ' + param.name + ', '
        call_expr += param.name + ', '
    if nargs > 0:
        ndecl = ndecl[:-2]
        call_expr = call_expr[:-2]
    ndecl += ')'
    call_expr += ')'
    ndecl += " { return %s; }" % call_expr
    return ndecl



from __future__ import unicode_literals

from numpy.testing import assert_almost_equal

from ncephes import misc


def test_lbeta():
    print(misc.lbeta(10, 3))
    assert_almost_equal(misc.lbeta(10, 3), -6.4922398350204711)



from six.moves.urllib.request import urlopen
import sys
import shutil
import os
from os import listdir
from os.path import join
from os.path import isdir
from os.path import splitext
import re
import tarfile
import subprocess
from build_helpers import get_supported_modules
from build_helpers import api_fdecls
from build_helpers import forward_call
from module_info import get_fdecls


SUPPORTED_MODULES = open('supported_modules.txt').read().split("\n")[:-1]


def download_extract():
    url = "http://www.netlib.org/cephes/"
    print("Downloading...")
    page = urlopen(url).read()

    c = re.compile(r'"([^"]+\.tgz)"')
    pkgs = re.findall(c, page)
    pkgs = [p for p in pkgs if splitext(p)[0] in SUPPORTED_MODULES]

    print("Extracting...")
    for pkg in pkgs:
        stream = urlopen(url + pkg)
        content = tarfile.open(fileobj=stream, mode="r|gz")
        name = splitext(pkg)[0]
        path = join('ncephes', 'cephes', name)
        if os.path.exists(path):
            shutil.rmtree(path)
        if name == 'c9x-complex':
            content.extractall(join('ncephes', 'cephes'))
        else:
            content.extractall(path)

    subprocess.check_call('chmod -R u+rw ncephes/cephes/', shell=True)
    subprocess.check_call('chmod -R go+r ncephes/cephes/', shell=True)


def clear_code():
    print("Cleaning files...")
    root = join('ncephes', 'cephes')
    dirs = [d for d in listdir(root) if isdir(join(root, d))]
    for d in dirs:
        for f in listdir(join(root, d)):
            fp = join(root, d, f)
            cmd = "perl -i -pe 's/[\x0c]//g' %s" % fp
            subprocess.check_call(cmd, shell=True)
            if fp.endswith('.c') or fp.endswith('.h'):
                cmd = "perl -i -pe 's/(ncephes_)*cabs/ncephes_cabs/g' %s" % fp
                subprocess.check_call(cmd, shell=True)
                cmd = "perl -i -pe 's/(ncephes_)*csqrt/ncephes_csqrt/g' %s" % fp
                subprocess.check_call(cmd, shell=True)
                if f != 'gamma.c' and f != 'incbet.c':
                    cmd = "perl -i -pe 's/(cephes_)*gamma/cephes_gamma/g' %s" % fp
                    subprocess.check_call(cmd, shell=True)


def ncephes_consts():
    print("Setting ncephes consts...")
    root = join('ncephes', 'cephes')
    dirs = [d for d in listdir(root) if isdir(join(root, d))]
    for d in dirs:
        for f in listdir(join(root, d)):
            fp = join(root, d, f)
            if fp.endswith('.c') or fp.endswith('.h'):
                cmd = "perl -i -pe 's/(NCEPHES_)*INFINITY/NCEPHES_INF/g' %s" % fp
                subprocess.check_call(cmd, shell=True)
                cmd = "perl -i -pe 's/(NCEPHES_)*NAN/NCEPHES_NAN/g' %s" % fp
                subprocess.check_call(cmd, shell=True)
                cmd = "perl -i -pe 's/(NCEPHES_)*MAXNUM/NCEPHES_MAXNUM/g' %s" % fp
                subprocess.check_call(cmd, shell=True)
                cmd = "perl -i -pe 's/(NCEPHES_)*PI/NCEPHES_PI/g' %s" % fp
                subprocess.check_call(cmd, shell=True)


def convert_old_style_proto():
    print("Converting K&R to ANSI style for function declaration...")
    root = join('ncephes', 'cephes')
    dirs = [d for d in listdir(root) if isdir(join(root, d))]
    for d in dirs:
        if d not in SUPPORTED_MODULES:
            continue
        for f in listdir(join(root, d)):
            if not (f.endswith('.c') or f.endswith('.h')):
                continue
            fp = join(root, d, f)
            incl1 = join(os.getcwd(), root, d)
            incl2 = join(os.getcwd(), root)
            cmd = "cproto -q -I%s -I%s -a %s" % (incl1, incl2, fp)
            subprocess.check_call(cmd, shell=True)


def _unlink(f):
    if os.path.exists(f):
        os.unlink(f)


def apply_patch():
    cmd = "patch ncephes/cephes/polyn/polyn.c ncephes/cephes/polyn.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/polyn/polyn.c ncephes/cephes/polyn.2.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/polyn/polmisc.c ncephes/cephes/polmisc.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/polyn/polmisc.c ncephes/cephes/polmisc.2.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/cprob/gamma.c ncephes/cephes/gamma.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/cprob/gamma.c ncephes/cephes/gamma.2.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/cprob/gamma.c ncephes/cephes/gamma.3.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/cprob/gamma.c ncephes/cephes/gamma.4.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/misc/beta.c ncephes/cephes/beta.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/misc/beta.c ncephes/cephes/beta.2.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/misc/beta.c ncephes/cephes/beta.3.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/cprob/incbet.c ncephes/cephes/incbet.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/ellf/ellf.c ncephes/cephes/ellf.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/misc/revers.c ncephes/cephes/revers.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/polyn/revers.c ncephes/cephes/revers.2.patch"
    subprocess.check_call(cmd, shell=True)
    cmd = "patch ncephes/cephes/polyn/revers.c ncephes/cephes/revers.3.patch"
    subprocess.check_call(cmd, shell=True)
    _unlink('ncephes/cephes/cmath/atan.c')
    _unlink('ncephes/cephes/cmath/exp.c')
    _unlink('ncephes/cephes/eval/protos.h')

    _unlink('ncephes/cephes/cprob/const.c')
    _unlink('ncephes/cephes/ellf/const.c')
    _unlink('ncephes/cephes/cmath/const.c')

    _unlink('ncephes/cephes/cprob/mconf.h')
    _unlink('ncephes/cephes/ellf/mconf.h')
    _unlink('ncephes/cephes/cmath/mconf.h')
    _unlink('ncephes/cephes/eval/mconf.h')

    _unlink('ncephes/cephes/polyn/polyr.c')

    _unlink('ncephes/cephes/cprob/mtherr.c')
    _unlink('ncephes/cephes/ellf/mtherr.c')
    _unlink('ncephes/cephes/cmath/mtherr.c')
    _unlink('ncephes/cephes/eval/mtherr.c')
    _unlink('ncephes/cephes/misc/mtherr.c')


def _create_api(module):
    h = module.upper() + '_H'
    fdecls = get_fdecls(module)
    apidecls = [api_fdecls(f) for f in fdecls]
    ffcalls = [forward_call(f) for f in fdecls]

    guard_start = "#ifndef %s\n#define %s\n\n" % (h, h)
    guard_end = "\n\n#endif\n"
    apidecls = '\n'.join(apidecls)
    with open(join('ncephes', 'include', 'ncephes', module + '.h'), 'w') as f:
        f.write(guard_start + apidecls + guard_end)

    with open(join('ncephes', 'cephes', module + '_ffcall.c'), 'w') as f:
        f.write('#include "ncephes/' + module + '.h"\n\n')
        if len(apidecls) == 0:
            f.write('int _do_nothing_%s() { return 0; }' % module)
        else:
            f.write('\n'.join(fdecls) + '\n\n')
            f.write('\n'.join(ffcalls) + '\n\n')


def create_api():
    modules = get_supported_modules()
    for module in modules:
        _create_api(module)


def update():
    src_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    old_path = os.getcwd()
    os.chdir(src_path)

    try:
        download_extract()
        clear_code()
        convert_old_style_proto()
        apply_patch()
        ncephes_consts()
        create_api()
    finally:
        os.chdir(old_path)

    print("Finished.")

if __name__ == '__main__':
    update()



from __future__ import unicode_literals

import os


def get_include():
    """
    Return the directory that contains the nCephes \\*.h header files.
    Extension modules that need to compile against nCephes should use this
    function to locate the appropriate include directory.
    Notes
    -----
    When using ``distutils``, for example in ``setup.py``.
    ::
        import ncephes as nc
        ...
        Extension('extension_name', ...
                include_dirs=[nc.get_include()])
        ...
    """
    import ncephes
    d = os.path.join(os.path.dirname(ncephes.__file__), 'include')
    return d


def get_lib():
    import ncephes
    d = os.path.join(os.path.dirname(ncephes.__file__), 'lib')
    return d



from __future__ import unicode_literals

from ncephes import cprob
from ncephes.cprob import incbet
from numpy.testing import assert_almost_equal

try:
    from numba import jit

    @jit(nogil=True, nopython=True)
    def _incbet(a, b, x):
        return incbet(a, b, x)

    @jit(nogil=True)
    def erf(x):
        return cprob.erf(x)

    def test_incbet():
        assert_almost_equal(_incbet(1., 3., 0.3), 0.657)

    def test_erf():
        assert_almost_equal(erf(1.3), 0.9340079449406524)

except ImportError:
    pass



from __future__ import unicode_literals

import os
from os.path import join
from tempfile import mkdtemp
from shutil import rmtree
from subprocess import check_output

from numpy.testing import assert_equal

from ncephes import get_include
from ncephes import get_lib


def test_get_include():
    suffix = os.path.join("ncephes", "include")
    assert_equal(get_include().endswith(suffix), True)


def test_get_lib():
    suffix = os.path.join("ncephes", "lib")
    assert_equal(get_lib().endswith(suffix), True)


def test_link_lib():
    folder = mkdtemp(dir='.')
    testc = join(folder, 'test.c')
    with open(testc, 'w') as f:
        f.write('''
#include <stdio.h>
#include "ncephes/cprob.h"

int main()
{
  printf("incbet: %.3f", ncephes_incbet(1., 3., 0.3));
  return 0;
}
''')
    from distutils.ccompiler import new_compiler
    from distutils import log
    compiler = new_compiler()
    objs = compiler.compile([testc], include_dirs=[get_include()])
    libraries = ['ncprob']
    if 'msv' not in compiler.__class__.__name__.lower():
        libraries += 'm'
    compiler.link_executable(objs, join(folder, 'test_link_lib'),
                             libraries=libraries,
                             library_dirs=[get_lib()])
    assert_equal(check_output(join(folder, 'test_link_lib'), shell=True),
                 b"incbet: 0.657")
    rmtree(folder)



from __future__ import absolute_import as _absolute_import

import _cffi_backend
from pkg_resources import DistributionNotFound as _DistributionNotFound
from pkg_resources import get_distribution as _get_distribution

from . import cprob, ellf
from .api import get_include, get_lib

try:
    __version__ = _get_distribution('ncephes').version
except _DistributionNotFound:
    __version__ = 'unknown'


def test():
    import os
    p = __import__('ncephes').__path__[0]
    src_path = os.path.abspath(p)
    old_path = os.getcwd()
    os.chdir(src_path)

    try:
        return_code = __import__('pytest').main(['-q'])
    finally:
        os.chdir(old_path)

    if return_code == 0:
        print("Congratulations. All tests have passed!")

    return return_code
